extern crate pancurses;
use self::pancurses::{
    initscr, endwin, noecho, has_colors, start_color, init_pair,
    Window,
    COLOR_GREEN, COLOR_BLACK
};

extern crate rand;
use self::rand::Rng;
use self::rand::prelude::SliceRandom;

use std::{thread, time};

use config::{Config};
use util::{copy_shuffle};

// Board control and interaction. The view & controller.

mod board {}

const DEFAULT_ROWS: usize = 17;
const DEFAULT_COLS: usize = 24;

// Different actions are printed to the console at different speeds (the terminal displaying output, vs the player "typing")
const TERMINAL_PRINTING_SPEED: time::Duration = time::Duration::from_millis(20);
const TYPING_SPEED: time::Duration = time::Duration::from_millis(70);

pub struct Screen {
    pub w: Window,
    attempts_left: u8,
    b: Board
}

struct Board {
    rows: usize,
    cols: usize,
    c: Vec<char>
}

struct Coord {
    start: usize,
    end: usize
}

impl Board {
    fn new(rows: Option<usize>, cols: Option<usize>) -> Board {
        let r = match rows {
            Some(r) => r,
            None => DEFAULT_ROWS
        };
        let c = match cols {
            Some(c) => c,
            None => DEFAULT_COLS
        };
        Board {
            rows: r,
            cols: c,
            c: Board::garbage_vec(r, c)
        }
    }

    fn garbage_vec(rows: usize, cols: usize) -> Vec<char> {
        let g: Vec<char> = "!@#$%^*()_-=+\\|/[]{}?\"\':;,.<>".chars().collect();
        let mut b = Vec::with_capacity(rows * cols);
        let mut rng = rand::thread_rng();

        for _ in 0..rows * cols {
            b.push(*g.choose(&mut rng).unwrap());
        }

        b
    }

    fn get(&self, i: usize) -> char {
        self.c[i]
    }

    fn getxy(&self, x: usize, y: usize) -> char {
        self.c[x*y]
    }

    fn set(&mut self, i: usize, c: char) {
        self.c[i] = c;
    }

    fn setxy(&mut self, x: usize, y: usize, c: char) {
        self.c[x*y] = c;
    }
}

impl Screen {
    pub fn new(conf: &Config) -> Screen {
        let window = initscr();
        noecho();
        window_enable_colors();
        Screen {
            w: window,
            attempts_left: 4,
            b: Board::new(conf.rows, conf.cols)
        }
    }

    pub fn intro(&self) {
        self.w.clear();
        thread::sleep(time::Duration::from_millis(250));
        self.print_with_delay(0, String::from("WELCOME TO ROBCO INDUSTRIES (TM) TERMLINK"), true, TERMINAL_PRINTING_SPEED);
        self.move_cursor_to(1, 0);
        thread::sleep(time::Duration::from_millis(30));

        // TODO the rest of the intro

    }

    pub fn initialize_game(&self, conf: Config) {
        self.print_with_delay(0, String::from("ROBCO INDUSTRIES (TM) TERMLINK PROTOCOL"), false, TERMINAL_PRINTING_SPEED);
        self.print_with_delay(1, String::from("ENTER PASSWORD NOW"), false, TERMINAL_PRINTING_SPEED);
        self.print_with_delay(
            3,
            self.attempts_display(),
            false,
            TERMINAL_PRINTING_SPEED
        );
        let border_hex = border_hex_gen(self.b.rows);

        let shuffled_words = copy_shuffle(&conf.words);

        let word_len = conf.words[0].len();
        // Assumes validate() ensured `words` is not empty, and each word is the same length
        let required_tiles = word_len * conf.words.len() + (conf.words.len() - 1);
        let board_tiles = self.b.rows * self.b.cols;
        if required_tiles > board_tiles {
            panic!("{} words of length {} requires {} tiles, but an {}x{} board only allows for {} tiles", conf.words.len(), word_len, required_tiles, self.b.cols, self.b.rows, board_tiles);
        }

        name_this(&self.b);
        // TODO: insert each word, of size N, into the array, with at least 1 character between each

    }

    pub fn end_window(&self) {
        endwin();
    }

    fn move_cursor_to(&self, y: i32, x: i32) {
        self.w.mv(y, x);
        self.w.refresh();
    }

    fn print_with_delay(&self, line: i32, string: String, skippable: bool, delay: time::Duration) {
        self.w.mv(line, 0);
        for c in string.chars() {
            self.w.addch(c);
            self.w.refresh();
            if skippable {
                // TODO keyboard skip
            }
            thread::sleep(delay);
        }
    }

    fn attempts_display(&self) -> String {
        format!("{} ATTEMPT(S) LEFT:{}", self.attempts_left, " *".repeat(self.attempts_left as usize))
    }

}

fn window_enable_colors() {
    if has_colors() {
        start_color();
        init_pair(1, COLOR_GREEN, COLOR_BLACK);
    }
}

fn border_hex_gen(rows: usize) -> Vec<u32> {
    // Build the hex values, printed alongside the text and garbage.
    // The vector will be of length rows * 2
    // TODO: this would make a nice generator, once those are stable.
    let mut rng = rand::thread_rng();
    let mut hex = rng.gen_range(0,200) + 63744;

    let mut ls = Vec::with_capacity(rows * 2);
    for i in 0..rows * 2 {
        ls[i] = hex;
        hex += 12;
    }

    ls
}

// Think of the board in word_len "chunks": place a word in each chunk. Then "shake" each one around a bit, in a random order, 3 times/word.
fn name_this(b: &Board) -> Vec<Coord> {
    // TODO
    return Vec::new();
}
